@@ -18,8 +18,10 @@ use openpgp::{
     packet::prelude::*,
     policy::Policy,
     types::*,
-    packet::key::*,
+    packet::{self, key::*},
     parse::{
+        PacketParser,
+        PacketParserResult,
         Parse,
         stream::*,
     },
@@ -63,49 +65,22 @@ pub fn cmd_decrypt(config: &crate::Config, args: &[String])
     let filename = args.get(0).cloned().unwrap_or_else(|| "-".into());
     let message = utils::open(config, &filename)?;
 
-    let policy = config.policy();
-
     // XXX: Currently, there is no nice way to disable armoring when
     // using the streaming decryptor.
 
-    let mut sink = if let Some(name) = config.outfile() {
+    let sink = if let Some(name) = config.outfile() {
         utils::create(config, name)?
     } else {
         Box::new(io::stdout())
     };
 
-    // Note: we use crypto::Decryptors backed by the gpg-agent.
-    // Currently, it is not safe to use these from async contexts,
-    // because they evaluate futures using a runtime, which may not be
-    // nested.  Therefore, the following code may not be run in an
-    // async context.
-    let transaction = || -> Result<()> {
-        let helper = DHelper::new(config, VHelper::new(config, 1));
-        let message = DecryptorBuilder::from_reader(message)?;
-        let mut d = match message.with_policy(policy, config.now(), helper) {
-            Ok(d) => d,
-            Err(e) => if config.list_only {
-                return Ok(());
-            } else {
-                return Err(e);
-            },
-        };
-
-        if ! config.list_only {
-            io::copy(&mut d, &mut sink)?;
-        }
-        let helper = d.into_helper();
-
-        if ! config.list_only {
-            helper.config.status().emit(Status::DecryptionOkay)?;
-            // For compatibility reasons we issue GOODMDC also for AEAD messages.
-            helper.config.status().emit(Status::GoodMDC)?;
-        }
-
-        Ok(())
+    let r = if config.unwrap_encryption {
+        decrypt_unwrap(config, Ok(message), Ok(sink))
+    } else {
+        decrypt(config, Ok(message), Ok(sink))
     };
 
-    match transaction() {
+    match r {
         Ok(()) => config.status().emit(Status::EndDecryption)?,
         Err(e) => {
             match e.downcast_ref::<openpgp::Error>() {
@@ -162,49 +137,23 @@ pub fn cmd_decrypt_files(config: &crate::Config, args: &[String])
         args
     };
 
-    let policy = config.policy();
     for ciphertext in inputs {
         config.status().emit(Status::FileStart {
             what: crate::status::FileStartOperation::Encrypt,
             name: &ciphertext,
         })?;
 
-        // Note: we use crypto::Decryptors backed by the gpg-agent.
-        // Currently, it is not safe to use these from async contexts,
-        // because they evaluate futures using a runtime, which may not be
-        // nested.  Therefore, the following code may not be run in an
-        // async context.
-        let transaction = || -> Result<()> {
-            let message = utils::open(config, &ciphertext)?;
-            let mut sink = utils::create(
-                config, &utils::make_outfile_name(ciphertext)?)?;
+        let message = utils::open(config, &ciphertext);
+        let sink = utils::make_outfile_name(ciphertext)
+            .and_then(|name| utils::create(config, &name));
 
-            let helper = DHelper::new(config, VHelper::new(config, 1));
-            let message = DecryptorBuilder::from_reader(message)?;
-            let mut d = match message.with_policy(policy, config.now(), helper) {
-                Ok(d) => d,
-                Err(e) => if config.list_only {
-                    return Ok(());
-                } else {
-                    return Err(e);
-                },
-            };
-
-            if ! config.list_only {
-                io::copy(&mut d, &mut sink)?;
-            }
-            let helper = d.into_helper();
-
-            if ! config.list_only {
-                helper.config.status().emit(Status::DecryptionOkay)?;
-                // For compatibility reasons we issue GOODMDC also for AEAD messages.
-                helper.config.status().emit(Status::GoodMDC)?;
-            }
-
-            Ok(())
+        let r = if config.unwrap_encryption {
+            decrypt_unwrap(config, message, sink)
+        } else {
+            decrypt(config, message, sink)
         };
 
-        if let Err(e) = transaction() {
+        if let Err(e) = r {
             config.error(format_args!("{}", e));
         }
         config.status().emit(Status::EndDecryption)?;
@@ -214,6 +163,111 @@ pub fn cmd_decrypt_files(config: &crate::Config, args: &[String])
     Ok(())
 }
 
+/// Decrypts `message` writing the plaintext to `sink`.
+fn decrypt(config: &crate::Config,
+           message: Result<Box<dyn io::Read + Send + Sync>>,
+           sink: Result<Box<dyn io::Write + Send + Sync>>)
+           -> Result<()>
+{
+    let message = message?;
+    let mut sink = sink?;
+
+    let helper = DHelper::new(config, VHelper::new(config, 1));
+    let message = DecryptorBuilder::from_reader(message)?;
+    let mut d = match message.with_policy(config.policy(), config.now(), helper) {
+        Ok(d) => d,
+        Err(e) => if config.list_only {
+            return Ok(());
+        } else {
+            return Err(e);
+        },
+    };
+
+    if ! config.list_only {
+        io::copy(&mut d, &mut sink)?;
+    }
+    let helper = d.into_helper();
+
+    if ! config.list_only {
+        helper.config.status().emit(Status::DecryptionOkay)?;
+        // For compatibility reasons we issue GOODMDC also for AEAD messages.
+        helper.config.status().emit(Status::GoodMDC)?;
+    }
+
+    Ok(())
+}
+
+/// Decrypts `message` writing the plaintext packet structure to
+/// `sink`.
+fn decrypt_unwrap(config: &crate::Config,
+                  message: Result<Box<dyn io::Read + Send + Sync>>,
+                  sink: Result<Box<dyn io::Write + Send + Sync>>)
+                  -> Result<()>
+{
+    let message = message?;
+    let mut sink = sink?;
+
+    let mut helper = DHelper::new(config, VHelper::new(config, 1));
+    let mut ppr = PacketParser::from_reader(message)?;
+
+    let mut pkesks: Vec<packet::PKESK> = Vec::new();
+    let mut skesks: Vec<packet::SKESK> = Vec::new();
+    while let PacketParserResult::Some(mut pp) = ppr {
+        helper.inspect(&pp)?;
+
+        let sym_algo_hint = if let Packet::AED(ref aed) = pp.packet {
+            Some(aed.symmetric_algo())
+        } else {
+            None
+        };
+
+        match pp.packet {
+            Packet::SEIP(_) | Packet::AED(_) => {
+                {
+                    let decrypt = |algo, secret: &SessionKey| {
+                        pp.decrypt(algo, secret).is_ok()
+                    };
+                    helper.decrypt(&pkesks[..], &skesks[..], sym_algo_hint,
+                                   decrypt)?;
+                }
+                if ! pp.processed() {
+                    return if config.list_only {
+                        Ok(())
+                    } else {
+                        Err(
+                            openpgp::Error::MissingSessionKey(
+                                "No session key".into()).into())
+                    };
+                }
+
+                io::copy(&mut pp, &mut sink)?;
+                break;
+            },
+            #[allow(deprecated)]
+            Packet::MDC(ref mdc) => if ! mdc.valid() {
+                return Err(openpgp::Error::ManipulatedMessage.into());
+            },
+            _ => (),
+        }
+
+        let (p, ppr_tmp) = pp.recurse()?;
+        match p {
+            Packet::PKESK(pkesk) => pkesks.push(pkesk),
+            Packet::SKESK(skesk) => skesks.push(skesk),
+            _ => (),
+        }
+        ppr = ppr_tmp;
+    }
+
+    if ! config.list_only {
+        helper.config.status().emit(Status::DecryptionOkay)?;
+        // For compatibility reasons we issue GOODMDC also for AEAD messages.
+        helper.config.status().emit(Status::GoodMDC)?;
+    }
+
+    Ok(())
+}
+
 pub struct DHelper<'a, 'store> {
     config: &'a crate::Config<'store>,
     vhelper: VHelper<'a, 'store>,
@@ -286,16 +340,15 @@ impl<'a, 'store> DHelper<'a, 'store> {
 
     /// Tries to decrypt the given PKESK packet with `keypair` and try
     /// to decrypt the packet parser using `decrypt`.
-    async fn try_decrypt<D>(&self,
-                            cert: &Cert,
-                            pkesk: &PKESK,
-                            sym_algo: Option<SymmetricAlgorithm>,
-                            mut keypair: KeyPair,
-                            decrypt: &mut D)
-                            -> Result<(Option<Fingerprint>,
-                                       SymmetricAlgorithm,
-                                       SessionKey)>
-        where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
+    async fn try_decrypt(&self,
+                         cert: &Cert,
+                         pkesk: &PKESK,
+                         sym_algo: Option<SymmetricAlgorithm>,
+                         mut keypair: KeyPair,
+                         decrypt: &mut dyn FnMut(SymmetricAlgorithm, &SessionKey) -> bool)
+                         -> Result<(Option<Fingerprint>,
+                                    SymmetricAlgorithm,
+                                    SessionKey)>
     {
         let kek = keypair.decrypt(pkesk.esk(),
                                   sym_algo.and_then(|a| a.key_size().ok()))?;
@@ -343,12 +396,10 @@ impl<'a, 'store> DHelper<'a, 'store> {
         }
     }
 
-    async fn async_decrypt<D>(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
-                              sym_algo: Option<SymmetricAlgorithm>,
-                              mut decrypt: D)
-                              -> Result<Option<openpgp::Fingerprint>>
-    where
-        D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool,
+    async fn async_decrypt(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
+                           sym_algo: Option<SymmetricAlgorithm>,
+                           decrypt: &mut dyn FnMut(SymmetricAlgorithm, &SessionKey) -> bool)
+                           -> Result<Option<openpgp::Fingerprint>>
     {
         // We provide all the information upfront before we try any
         // decryption.
@@ -384,7 +435,10 @@ impl<'a, 'store> DHelper<'a, 'store> {
         // First, try public key encryption.
         let mut success = None;
         let mut pkesks_results: Vec<(&PKESK, Option<error_codes::Error>)> =
-            pkesks.into_iter().map(|p| (p, None)).collect();
+            pkesks.into_iter()
+            .filter(|p| self.config.policy.public_key_algorithm(p.pk_algo()).is_ok())
+            .map(|p| (p, None)).collect();
+
         for (pkesk, error) in pkesks_results.iter_mut() {
             let keyid = pkesk.recipient();
             let handle = KeyHandle::from(keyid);
@@ -423,6 +477,8 @@ impl<'a, 'store> DHelper<'a, 'store> {
             let cert = match self.config.keydb().lookup_by_cert_or_subkey(&handle).ok()
                 .and_then(|c| c.into_iter().next())
                 .and_then(|c| c.to_cert().ok().cloned())
+                .filter(|c| self.config.policy.public_key_algorithm(
+                    c.primary_key().pk_algo()).is_ok())
             {
                 Some(c) => c,
                 None => {
@@ -488,7 +544,7 @@ impl<'a, 'store> DHelper<'a, 'store> {
             }
 
             if let Ok(r) = self.try_decrypt(
-                &cert, pkesk, sym_algo, pair, &mut decrypt).await
+                &cert, pkesk, sym_algo, pair, decrypt).await
             {
                 // Success!
                 success = Some(r);
@@ -712,12 +768,13 @@ impl<'a, 'store> DHelper<'a, 'store> {
 impl<'a, 'store> DecryptionHelper for DHelper<'a, 'store> {
     fn decrypt<D>(&mut self, pkesks: &[PKESK], skesks: &[SKESK],
                   sym_algo: Option<SymmetricAlgorithm>,
-                  decrypt: D) -> Result<Option<openpgp::Fingerprint>>
+                  mut decrypt: D) -> Result<Option<openpgp::Fingerprint>>
         where D: FnMut(SymmetricAlgorithm, &SessionKey) -> bool
     {
         let rt = tokio::runtime::Runtime::new()?;
         let r =
-            rt.block_on(self.async_decrypt(pkesks, skesks, sym_algo, decrypt));
+            rt.block_on(self.async_decrypt(pkesks, skesks, sym_algo,
+                                           &mut decrypt));
 
         if r.is_err() && ! self.config.list_only {
             self.config.status().emit(Status::DecryptionFailed)?;
