@@ -43,7 +43,7 @@ mod macros;
 pub mod tracing;
 #[macro_use]
 pub mod argparse;
-use argparse::{Argument, Opt, flags::*};
+use argparse::Argument;
 pub mod babel;
 pub mod clock;
 pub mod common;
@@ -79,396 +79,8 @@ pub mod generate_key;
 pub mod filter;
 pub mod quick;
 
-/// Commands and options.
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-#[allow(non_camel_case_types)]
-pub enum CmdOrOpt {
-    aNull = 0,
-    o1 = 1,
-    oArmor	  = 'a' as isize,
-    aDetachedSign = 'b' as isize,
-    aSym	  = 'c' as isize,
-    aDecrypt	  = 'd' as isize,
-    aEncr	  = 'e' as isize,
-    oRecipientFile       = 'f' as isize,
-    oHiddenRecipientFile = 'F' as isize,
-    oInteractive  = 'i' as isize,
-    aListKeys	  = 'k' as isize,
-    oDryRun	  = 'n' as isize,
-    oOutput	  = 'o' as isize,
-    oQuiet	  = 'q' as isize,
-    oRecipient	  = 'r' as isize,
-    oHiddenRecipient = 'R' as isize,
-    aSign	  = 's' as isize,
-    oTextmodeShort= 't' as isize,
-    oLocalUser	  = 'u' as isize,
-    oVerbose	  = 'v' as isize,
-    oCompress	  = 'z' as isize,
-    oSetNotation  = 'N' as isize,
-    aListSecretKeys = 'K' as isize,
-    o300 = 300,
-    o301,
-    o302,
-    o303,
-    o309 = 309,
-    oBatch	  = 500,
-    oMaxOutput,
-    oInputSizeHint,
-    oSigNotation,
-    oCertNotation,
-    oShowNotation,
-    oNoShowNotation,
-    oKnownNotation,
-    aEncrFiles,
-    aEncrSym,
-    aDecryptFiles,
-    aClearsign,
-    aStore,
-    aQuickKeygen,
-    aFullKeygen,
-    aKeygen,
-    aSignEncr,
-    aSignEncrSym,
-    aSignSym,
-    aSignKey,
-    aLSignKey,
-    aQuickSignKey,
-    aQuickLSignKey,
-    aQuickRevSig,
-    aQuickAddUid,
-    aQuickAddKey,
-    aQuickRevUid,
-    aQuickSetExpire,
-    aQuickSetPrimaryUid,
-    aListConfig,
-    aListGcryptConfig,
-    aGPGConfList,
-    aGPGConfTest,
-    aListPackets,
-    aEditKey,
-    aDeleteKeys,
-    aDeleteSecretKeys,
-    aDeleteSecretAndPublicKeys,
-    aImport,
-    aFastImport,
-    aVerify,
-    aVerifyFiles,
-    aListSigs,
-    aSendKeys,
-    aRecvKeys,
-    aLocateKeys,
-    aLocateExtKeys,
-    aSearchKeys,
-    aRefreshKeys,
-    aFetchKeys,
-    aShowKeys,
-    aExport,
-    aExportSecret,
-    aExportSecretSub,
-    aExportSshKey,
-    aCheckKeys,
-    aGenRevoke,
-    aDesigRevoke,
-    aPrimegen,
-    aPrintMD,
-    aPrintMDs,
-    aCheckTrustDB,
-    aUpdateTrustDB,
-    aFixTrustDB,
-    aListTrustDB,
-    aListTrustPath,
-    aExportOwnerTrust,
-    aImportOwnerTrust,
-    aDeArmor,
-    aEnArmor,
-    aGenRandom,
-    aRebuildKeydbCaches,
-    aCardStatus,
-    aCardEdit,
-    aChangePIN,
-    aPasswd,
-    aServer,
-    aTOFUPolicy,
-
-    oMimemode,
-    oTextmode,
-    oNoTextmode,
-    oExpert,
-    oNoExpert,
-    oDefSigExpire,
-    oAskSigExpire,
-    oNoAskSigExpire,
-    oDefCertExpire,
-    oAskCertExpire,
-    oNoAskCertExpire,
-    oDefCertLevel,
-    oMinCertLevel,
-    oAskCertLevel,
-    oNoAskCertLevel,
-    oFingerprint,
-    oWithFingerprint,
-    oWithSubkeyFingerprint,
-    oWithICAOSpelling,
-    oWithKeygrip,
-    oWithSecret,
-    oWithWKDHash,
-    oWithColons,
-    oWithKeyData,
-    oWithKeyOrigin,
-    oWithTofuInfo,
-    oWithSigList,
-    oWithSigCheck,
-    oAnswerYes,
-    oAnswerNo,
-    oKeyring,
-    oPrimaryKeyring,
-    oSecretKeyring,
-    oShowKeyring,
-    oDefaultKey,
-    oDefRecipient,
-    oDefRecipientSelf,
-    oNoDefRecipient,
-    oTrySecretKey,
-    oOptions,
-    oDebug,
-    oDebugLevel,
-    oDebugAll,
-    oDebugIOLBF,
-    oStatusFD,
-    oStatusFile,
-    oAttributeFD,
-    oAttributeFile,
-    oEmitVersion,
-    oNoEmitVersion,
-    oCompletesNeeded,
-    oMarginalsNeeded,
-    oMaxCertDepth,
-    oLoadExtension,
-    oCompliance,
-    oGnuPG,
-    oRFC2440,
-    oRFC4880,
-    oRFC4880bis,
-    oOpenPGP,
-    oPGP6,
-    oPGP7,
-    oPGP8,
-    oDE_VS,
-    oMinRSALength,
-    oRFC2440Text,
-    oNoRFC2440Text,
-    oCipherAlgo,
-    oDigestAlgo,
-    oCertDigestAlgo,
-    oCompressAlgo,
-    oCompressLevel,
-    oBZ2CompressLevel,
-    oBZ2DecompressLowmem,
-    oPassphrase,
-    oPassphraseFD,
-    oPassphraseFile,
-    oPassphraseRepeat,
-    oPinentryMode,
-    oCommandFD,
-    oCommandFile,
-    oQuickRandom,
-    oNoVerbose,
-    oTrustDBName,
-    oNoSecmemWarn,
-    oRequireSecmem,
-    oNoRequireSecmem,
-    oNoPermissionWarn,
-    oNoArmor,
-    oNoDefKeyring,
-    oNoKeyring,
-    oNoGreeting,
-    oNoTTY,
-    oNoOptions,
-    oNoBatch,
-    oHomedir,
-    oSkipVerify,
-    oSkipHiddenRecipients,
-    oNoSkipHiddenRecipients,
-    oAlwaysTrust,
-    oTrustModel,
-    oForceOwnertrust,
-    oSetFilename,
-    oForYourEyesOnly,
-    oNoForYourEyesOnly,
-    oSetPolicyURL,
-    oSigPolicyURL,
-    oCertPolicyURL,
-    oShowPolicyURL,
-    oNoShowPolicyURL,
-    oSigKeyserverURL,
-    oUseEmbeddedFilename,
-    oNoUseEmbeddedFilename,
-    oComment,
-    oDefaultComment,
-    oNoComments,
-    oThrowKeyids,
-    oNoThrowKeyids,
-    oShowPhotos,
-    oNoShowPhotos,
-    oPhotoViewer,
-    oS2KMode,
-    oS2KDigest,
-    oS2KCipher,
-    oS2KCount,
-    oDisplayCharset,
-    oNotDashEscaped,
-    oEscapeFrom,
-    oNoEscapeFrom,
-    oLockOnce,
-    oLockMultiple,
-    oLockNever,
-    oKeyServer,
-    oKeyServerOptions,
-    oImportOptions,
-    oImportFilter,
-    oExportOptions,
-    oExportFilter,
-    oListOptions,
-    oVerifyOptions,
-    oTempDir,
-    oExecPath,
-    oEncryptTo,
-    oHiddenEncryptTo,
-    oNoEncryptTo,
-    oEncryptToDefaultKey,
-    oLoggerFD,
-    oLoggerFile,
-    oUtf8Strings,
-    oNoUtf8Strings,
-    oDisableCipherAlgo,
-    oDisablePubkeyAlgo,
-    oAllowNonSelfsignedUID,
-    oNoAllowNonSelfsignedUID,
-    oAllowFreeformUID,
-    oNoAllowFreeformUID,
-    oAllowSecretKeyImport,
-    oEnableSpecialFilenames,
-    oNoLiteral,
-    oSetFilesize,
-    oHonorHttpProxy,
-    oFastListMode,
-    oListOnly,
-    oIgnoreTimeConflict,
-    oIgnoreValidFrom,
-    oIgnoreCrcError,
-    oIgnoreMDCError,
-    oShowSessionKey,
-    oOverrideSessionKey,
-    oOverrideSessionKeyFD,
-    oOverrideComplianceCheck,
-    oNoRandomSeedFile,
-    oAutoKeyRetrieve,
-    oNoAutoKeyRetrieve,
-    oAutoKeyImport,
-    oNoAutoKeyImport,
-    oUseAgent,
-    oNoUseAgent,
-    oGpgAgentInfo,
-    oMergeOnly,
-    oTryAllSecrets,
-    oTrustedKey,
-    oNoExpensiveTrustChecks,
-    oFixedListMode,
-    oLegacyListMode,
-    oNoSigCache,
-    oAutoCheckTrustDB,
-    oNoAutoCheckTrustDB,
-    oPreservePermissions,
-    oDefaultPreferenceList,
-    oDefaultKeyserverURL,
-    oPersonalCipherPreferences,
-    oPersonalDigestPreferences,
-    oPersonalCompressPreferences,
-    oAgentProgram,
-    oDirmngrProgram,
-    oDisableDirmngr,
-    oDisplay,
-    oTTYname,
-    oTTYtype,
-    oLCctype,
-    oLCmessages,
-    oXauthority,
-    oGroup,
-    oUnGroup,
-    oNoGroups,
-    oStrict,
-    oNoStrict,
-    oMangleDosFilenames,
-    oNoMangleDosFilenames,
-    oEnableProgressFilter,
-    oMultifile,
-    oKeyidFormat,
-    oExitOnStatusWriteError,
-    oLimitCardInsertTries,
-    oReaderPort,
-    octapiDriver,
-    opcscDriver,
-    oDisableCCID,
-    oRequireCrossCert,
-    oNoRequireCrossCert,
-    oAutoKeyLocate,
-    oNoAutoKeyLocate,
-    oAllowMultisigVerification,
-    oEnableLargeRSA,
-    oDisableLargeRSA,
-    oEnableDSA2,
-    oDisableDSA2,
-    oAllowMultipleMessages,
-    oNoAllowMultipleMessages,
-    oAllowWeakDigestAlgos,
-    oAllowWeakKeySignatures,
-    oFakedSystemTime,
-    oNoAutostart,
-    oPrintPKARecords,
-    oPrintDANERecords,
-    oTOFUDefaultPolicy,
-    oTOFUDBFormat,
-    oDefaultNewKeyAlgo,
-    oWeakDigest,
-    oUnwrap,
-    oOnlySignTextIDs,
-    oDisableSignerUID,
-    oSender,
-    oKeyOrigin,
-    oRequestOrigin,
-    oNoSymkeyCache,
-    oUseOnlyOpenPGPCard,
-    oIncludeKeyBlock,
-    oNoIncludeKeyBlock,
-    oForceSignKey,
-    oForbidGenKey,
-    oRequireCompliance,
-
-    oNoop,
-
-    // Our own extensions.
-    aXSequoiaParcimonie,
-    oXSequoiaAutostartParcimonie,
-    aXSequoiaParcimonieDaemonize,
-    oXSequoiaQuietFakedSystemTime,
-
-    // Special, implicit commands.
-    aHelp = 'h' as isize,
-    aVersion = 32769,
-    aWarranty = 32770,
-    aDumpOptions = 32771,
-    aDumpOpttbl = 32772,
-}
-
-impl From<CmdOrOpt> for isize {
-    fn from(c: CmdOrOpt) -> isize {
-        c as isize
-    }
-}
-
-use CmdOrOpt::*;
-
-include!("gpg.option.inc");
+pub mod gpg_args;
+use gpg_args::CmdOrOpt;
 
 pub struct Config<'store> {
     // Runtime.
@@ -1392,6 +1004,8 @@ impl Preferences {
 
 fn set_cmd(cmd: &mut Option<CmdOrOpt>, new_cmd: CmdOrOpt)
            -> anyhow::Result<()> {
+    use CmdOrOpt::*;
+
     match cmd.as_ref().clone() {
         None => *cmd = Some(new_cmd),
         Some(c) if *c == new_cmd => (),
@@ -1617,6 +1231,8 @@ fn print_additional_version(config: &Config) {
 }
 
 fn real_main() -> anyhow::Result<()> {
+    use CmdOrOpt::*;
+
     tracing::parse_command_line();
 
     let parser = argparse::Parser::new(
@@ -1624,7 +1240,7 @@ fn real_main() -> anyhow::Result<()> {
         crate::gnupg_interface::VERSION,
         "Sign, check, encrypt or decrypt\n\
          Default operation depends on the input data",
-        &OPTIONS)
+        &gpg_args::OPTIONS)
         .with_additional_version_information(print_additional_version);
 
     let mut opt = Config::new()?;
@@ -1664,7 +1280,7 @@ fn real_main() -> anyhow::Result<()> {
             rarg.context("Error parsing command-line arguments")?;
         match arg {
             Argument::Option(aHelp, _) =>
-                return Ok(parser.help(&opt)),
+                return Ok(parser.help(&opt, sequoia_openpgp::VERSION)),
             Argument::Option(aWarranty, _) =>
                 return Ok(parser.warranty()),
             Argument::Option(aDumpOptions, _) =>
@@ -1703,7 +1319,6 @@ fn real_main() -> anyhow::Result<()> {
             },
         };
       let mut handle_argument = || -> Result<bool> {
-        use CmdOrOpt::*;
         match cmd {
 	    aListConfig
 	        | aListGcryptConfig
@@ -2368,7 +1983,7 @@ fn real_main() -> anyhow::Result<()> {
 	    },
 	    oSender => {
                 let sender = value.as_str().unwrap();
-                if let Some(v) = argparse::utils::mailbox_from_userid(sender)? {
+                if let Some(v) = utils::mailbox_from_userid(sender)? {
                     opt.sender_list.push(v);
                 } else {
                     return Err(anyhow::anyhow!(
@@ -2659,7 +2274,7 @@ fn real_main() -> anyhow::Result<()> {
     }
 
     if command == Some(aVersion) {
-        return Ok(parser.version(&opt));
+        return Ok(parser.version(&opt, sequoia_openpgp::VERSION));
     }
 
     if greeting && ! no_greeting {
@@ -2712,16 +2327,16 @@ fn real_main() -> anyhow::Result<()> {
 
     // Get the default one if no keyring has been specified.
     if keyrings.is_empty() {
-        opt.keydb.add_resource(&opt.homedir, "pubring.gpg", false, true)?;
+        opt.keydb.add_resource(&opt.homedir, "pubring.gpg", Some(true), false, true)?;
 
         // Use the keybox database if it is there.
         let _ = opt.keydb.add_resource(
             &opt.homedir, "gnupg-kbx-db:public-keys.d/pubring.db",
-            true, false)?;
+            Some(true), true, false)?;
     }
 
     for path in keyrings {
-        opt.keydb.add_resource(&opt.homedir, path, true, false)?;
+        opt.keydb.add_resource(&opt.homedir, path.as_ref(), None, true, false)?;
     }
 
     if let Some(aGPGConfTest) = command {
