@@ -37,36 +37,60 @@ use crate::{
     gnupg_interface::STRICT_OUTPUT,
 };
 
-pub struct Fd(Option<Mutex<RefCell<Box<dyn io::Write + Send + Sync>>>>);
+pub struct Fd {
+    stream: Option<Mutex<RefCell<Box<dyn io::Write + Send + Sync>>>>,
+    exit_on_write_error: bool,
+}
 
 impl<S: io::Write + Send + Sync + 'static> From<S> for Fd {
     fn from(s: S) -> Fd {
-        Fd(Some(Mutex::new(RefCell::new(Box::new(s)))))
+        Fd {
+            stream: Some(Mutex::new(RefCell::new(Box::new(s)))),
+            exit_on_write_error: false,
+        }
     }
 }
 
 impl Fd {
     /// Sends all output to /dev/null.
     pub fn sink() -> Self {
-        Fd(None)
+        Fd {
+            stream: None,
+            exit_on_write_error: false,
+        }
     }
 
     /// Whether gpg was called with --status-fd.
     ///
     /// If not, everything is sent to /dev/null.
     pub fn enabled(&self) -> bool {
-        self.0.is_some()
+        self.stream.is_some()
+    }
+
+    /// Replaces the current status-fd stream.
+    pub fn set_stream(&mut self, s: Box<dyn Write + Send + Sync>) {
+        self.stream = Some(Mutex::new(RefCell::new(s)));
+    }
+
+    /// Terminates the program when writing to the status-fd fails.
+    pub fn exit_on_write_error(&mut self) {
+        self.exit_on_write_error = true;
     }
 
     /// Emits a status message.
     #[allow(dead_code)]
     pub fn emit(&self, status: Status<'_>) -> Result<()> {
         crate::with_invocation_log(|sink| status.emit(sink));
-        if let Some(fd) = self.0.as_ref() {
-            status.emit(&mut *fd.lock().expect("not poisoned").borrow_mut())
-        } else {
-            Ok(())
+        if let Some(fd) = self.stream.as_ref() {
+            let r = status.emit(
+                &mut *fd.lock().expect("not poisoned").borrow_mut());
+
+            if self.exit_on_write_error && r.is_err() {
+                std::process::exit(0);
+            }
         }
+
+        Ok(())
     }
 
     /// Emits a status message or an interactive prompt.
@@ -79,7 +103,10 @@ impl Fd {
         if self.enabled() {
             self.emit(status)
         } else {
-            Ok(writeln!(io::stdout(), "{}", msg)?)
+            let mut sink = io::stdout();
+            write!(&mut sink, "{} ", msg)?;
+            sink.flush()?;
+            Ok(())
         }
     }
 }
@@ -135,12 +162,6 @@ pub enum Status<'a> {
         creation_time: SystemTime,
     },
 
-    AssertPubkeyAlgo {
-        fp: Fingerprint,
-        accepted: bool,
-        policy: String,
-    },
-
     // Encryption-related.
     EncTo {
         pkesk: &'a PKESK,
@@ -475,17 +496,6 @@ impl Status<'_> {
                          t.format("%s"))?;
             },
 
-            AssertPubkeyAlgo {
-                fp,
-                accepted,
-                policy,
-            } => {
-                writeln!(w, "ASSERT_PUBKEY_ALGO {:X} {} {}",
-                         fp,
-                         if *accepted { 1 } else { 0 },
-                         policy)?;
-            },
-
             EncTo {
                 pkesk,
             } => {
