@@ -18,7 +18,7 @@ use openpgp::{
     packet::signature::subpacket::*,
     packet::key,
     types::*,
-    policy::HashAlgoSecurity,
+    policy::{HashAlgoSecurity, Policy},
     serialize::Marshal,
 };
 use openpgp::cert::prelude::*;
@@ -220,7 +220,7 @@ pub fn cmd_verify(control: &crate::Config, args: &[String])
     // detached signature.
     let do_it = || -> Result<()> {
         if args.len() > 1 {
-            let data = utils::open_multiple(control, &args[1..]);
+            let data = utils::open_multiple(control, &args[1..])?;
             let helper = VHelper::new(control, 1);
             let mut v = DetachedVerifierBuilder::from_reader(sig)?
                 .with_policy(policy, control.now(), helper)?;
@@ -678,10 +678,6 @@ impl<'a, 'store> VHelper<'a, 'store> {
             primary: ka.cert().fingerprint(),
         })?;
 
-        // Check that the issuing key matches the policy (for
-        // --assert-pubkey-algo).
-        self.control.pubkey_algo_policy.check(self.control, ka.key())?;
-
         let validity =
             self.control.lookup_certs(
                 &Query::ExactKey(ka.cert().key_handle()))?
@@ -915,7 +911,6 @@ impl<'a, 'store> VHelper<'a, 'store> {
                         continue;
                     }
 
-                    let mut sig = (*sig).clone();
                     let openpgp_error = error.downcast_ref::<openpgp::Error>();
                     if sig.verify(ka.key()).is_ok()
                         && self.control.policy().signature(
@@ -990,6 +985,11 @@ impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
         Ok(ids.iter().filter_map(|id| self.control.keydb().lookup_by_cert_or_subkey(id).ok())
            .flatten()
            .filter_map(|cert| cert.to_cert().ok().cloned())
+           .filter(|c| self.control.policy.public_key_algorithm(
+               c.primary_key().pk_algo()).is_ok())
+           .map(|c| c.retain_subkeys(
+               |sk| self.control.policy.public_key_algorithm(
+                   sk.pk_algo()).is_ok()))
            .collect())
     }
 
